#ifndef MPU_INVERSE_SIGMA0_H
#define MPU_INVERSE_SIGMA0_H

#include "ptypes.h"

extern UV inverse_sigma0_count(UV lo, UV hi, UV k);
extern UV* inverse_sigma0_list(UV *count, UV lo, UV hi, UV k);

#endif
