#ifndef MPU_LEHMER_H
#define MPU_LEHMER_H

#include "ptypes.h"

extern UV legendre_prime_count(UV n);
extern UV meissel_prime_count(UV n);
extern UV lehmer_prime_count(UV n);
extern UV LMOS_prime_count(UV n);

#endif
